from zope.interface import Interface

# TODO: These interfaces are nowhere nearly as powerful as Clojure's
# protocols (e.g. no reify). I'm not sure they are even worth it
# except for maybe documentation.

class IChannel(Interface):

    def take(self, handler):
        pass

    def put(self, value, handler):
        pass

    def close(self):
        pass


class IHandler(Interface):

    def is_active(self):
        pass

    def commit(self):
        pass


class IBuffer(Interface):

    def is_full(self):
        pass

    def remove(self):
        pass

    def add(self, item):
        pass



from zope.interface import implements
from collections import namedtuple

from csp.impl import dispatch
from csp.impl.interfaces import IChannel
from csp.impl.buffers import RingBuffer, FixedBuffer


# The number of pending puts/takes between "garbage collections"
# (removing stale (no longer interested, inactive) handlers (e.g.
# unchosen alt operations))
MAX_DIRTY = 64

# Maximum pending puts/takes allowed on a single channel
MAX_QUEUE_SIZE = 1024

PutBox = namedtuple("PutBox", ["handler", "value"])

Box = namedtuple("Box", ["value"])

# TODO: 2 buffers can be replaced with 1 buffer + 1 flag: pending
# takes/puts. Should we do that?


class ManyToManyChannel:
    """Creates a channel with an optional buffer. If buf_or_n is a number,
    a fixed buffer of that size is created and used.
    """

    implements(IChannel)

    def __init__(self, buf_or_n = None):
        if buf_or_n == 0:
            buf_or_n = None
        if isinstance(buf_or_n, int):
            self.buf = FixedBuffer(buf_or_n)
        else:
            self.buf = buf_or_n

        self.closed = False
        self.takes = RingBuffer(32)
        self.puts = RingBuffer(32)
        # Counters towards next "garbage collections"
        self.dirty_takes = 0
        self.dirty_puts = 0

    def put(self, value, handler):
        if value is None:
            raise Exception("Cannot put None on a channel.")

        if self.closed or not handler.is_active():
            return Box(not self.closed)

        while True:
            try:
                taker = self.takes.pop()
            except IndexError:
                taker = None
            # There's a pending take...
            if taker is not None:
                # ... that is still interested
                if taker.is_active():
                    # Done, shake hand, here it is
                    callback, _ = taker.commit(), handler.commit()
                    # FIX
                    dispatch.run(lambda: callback(value))
                    return Box(True)
                else:
                    continue
            # No pending takes
            else:
                # Throw the value on the queue and be done with it
                if self.buf is not None and not self.buf.is_full():
                    handler.commit()
                    self.buf.add(value)
                    return Box(True)
                # No more room for waiting
                else:
                    # Periodically remove stale puts
                    if self.dirty_puts > MAX_DIRTY:
                        self.puts.cleanup(keep = lambda putter: putter.handler.is_active())
                        self.dirty_puts = 0
                    else:
                        self.dirty_puts += 1

                    # TODO: Do a last-chance "garbage collection" to
                    # be sure?
                    if len(self.puts) >= MAX_QUEUE_SIZE:
                        raise Exception("No more than %d pending puts are allowed on a single channel." % MAX_QUEUE_SIZE)

                    # Queue this put
                    self.puts.unbounded_unshift(PutBox(handler, value))
            break

    def take(self, handler):
        if not handler.is_active():
            return

        # Waiting values go first
        if self.buf is not None and len(self.buf) > 0:
            handler.commit()
            return Box(self.buf.remove())

        while True:
            try:
                putter = self.puts.pop()
            except IndexError:
                putter = None
            # There's a pending put...
            if putter is not None:
                put_handler = putter.handler
                # ... that is still interested
                if put_handler.is_active():
                    # Done, shake hand, take it
                    callback, _ = put_handler.commit(), handler.commit()
                    dispatch.run(lambda: callback(True))
                    return Box(putter.value)
                else:
                    continue
            # No pending puts
            else:
                if self.closed:
                    handler.commit()
                    return Box(None)
                else:
                    # Periodically remove stale takes
                    if self.dirty_takes > MAX_DIRTY:
                        self.takes.cleanup(keep = lambda handler: handler.is_active())
                        self.dirty_takes = 0
                    else:
                        self.dirty_takes += 1

                    # TODO: Do a last-chance "garbage collection" to
                    # be sure?
                    if len(self.takes) >= MAX_QUEUE_SIZE:
                        raise Exception("No more than %d pending takes are allowed on a single channel." % MAX_QUEUE_SIZE)

                    # Queue this take
                    self.takes.unbounded_unshift(handler)
            break

    def close(self):
        if self.closed:
            return

        # The current semantics is that a "close" is forceful, which
        # means any pending puts or buffered values will be ignored.
        # Is that actually a good thing? Ignoring buffered values
        # sounds ok, but no response to pending puts?

        self.closed = True
        while True:
            try:
                taker = self.takes.pop()
            except IndexError:
                break
            if taker.is_active():
                callback = taker.commit()
                dispatch.run(lambda: callback(None))



from csp import go, wait, take, stop


def lazy_echo(x):
    yield wait(0.1)
    print "I'm done"
    yield stop(x)


def main():
    chan = go(lazy_echo(1), True)
    print (yield take(chan))

    chan = go(lazy_echo(2), True)
    yield wait(1)
    print (yield take(chan))

    yield stop("Done")



from csp import Channel, alts, timeout, put_then_callback, no_op

from twisted.web.client import getPage


def request(url):
    channel = Channel(1)
    def ok(value):
        put_then_callback(channel, (value, None), no_op)
    def error(failure):
        put_then_callback(channel, (None, failure), no_op)
    getPage(url).addCallback(ok).addErrback(error)
    return channel


def main():
    for url, dt in (
        ("http://www.google.com/search?q=csp", 0.01), # too little time
        ("http://www.google.come/search?q=csp", 10),  # wrong domain name
        ("http://www.google.com/search?q=csp", 10),
    ):
        print "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
        print url, dt, "seconds"
        c = request(url)
        t = timeout(dt)
        value, chan = yield alts([c, t])
        if chan is c:
            result, error = value
            if error:
                print "Uhm, not good"
                print error
            else:
                print "Here"
                print result[:500] + "..."
        elif chan is t:
            print "Timeout"



# API
from csp.impl.buffers import FixedBuffer, DroppingBuffer, SlidingBuffer
from csp.impl.channels import ManyToManyChannel as Channel
from csp.impl.process import put_then_callback, take_then_callback
from csp.impl.process import put, take, wait, alts, stop
from csp.impl.timers import timeout

import csp.impl.process


def no_op(*arg):
    pass


def go(gen, chan=False):
    """Asynchronously executes a "go" function (a generator returned by a
    function following certain conventions, to be exact).

    Returns a channel that will receive the result of the go function
    when it completes.
    """
    # Only deliver the result to a channel if requested. Ideally I
    # think the choice should be a call-site optimization (whether the
    # return value of "go" is used). However that probably requires
    # the construct to be built-in (I believe Clojure's core.async
    # suffers from the same problem).
    if chan:
        channel = Channel(1)
        def done(value):
            if value is not None:
                put_then_callback(channel, value, no_op)
            channel.close()

        process = csp.impl.process.Process(gen, done)
        process.run()
        return channel
    else:
        process = csp.impl.process.Process(gen)
        process.run()
        return


# For API consistency (sort of)
def close(channel):
    """Closes a channel.

    - Pending puts are ignored.
    - Pending takes are flushed with None.
    - Future puts succeed immediately.
    - Future takes receive immediately None.
    """
    return channel.close()



# http://talks.golang.org/2012/concurrency.slide#34
# Fan-in using select

from csp import Channel, put, take, go, alts

from .boring import boring

def fan_in(input1, input2):
    c = Channel()
    def collect():
        while True:
            value, _ = yield alts([input1, input2])
            yield put(c, value)
    go(collect())
    return c


def main():
    c = fan_in(boring("Joe"), boring("Ann"))
    for i in range(10):
        print (yield take(c))
    print "You are boring; I'm leaving."



# http://talks.golang.org/2012/concurrency.slide#26
# Channels as a handle on a service

from csp import take

from .boring import boring


def main():
    joe = boring("Joe")
    ann = boring("Ann")
    for i in range(5):
        print (yield take(joe))
        print (yield take(ann))
    print "You are boring; I'm leaving."



# http://talks.golang.org/2012/concurrency.slide#30

from csp import Channel, put, take
from collections import namedtuple

Message = namedtuple("Message", ["str", "wait"])

for i in range(5):
    msg1 = yield take(chan)
    print msg1.str
    msg2 = yield take(chan)
    print msg2.str
    yield put(msg1.wait, True)
    yield put(msg2.wait, True)

wait_for_it = Channel()

yield put(c,)



# http://talks.golang.org/2012/concurrency.slide#25
# Generator: function that returns a channel

from csp import Channel, put, take, go, wait
import random


def boring(message):
    c = Channel()
    def _do():
        i = 0
        while True:
            yield put(c, "%s %d" % (message, i))
            yield wait(random.random())
            i += 1
    go(_do())
    return c


def main():
    b = boring("boring!")
    for i in range(5):
        print "You say: \"%s\"" % (yield take(b))
    print "You are boring; I'm leaving."



# http://talks.golang.org/2012/concurrency.slide#35
# Timeout using select

from csp import alts, timeout, stop

from .boring import boring

def main():
    c = boring("Joe")
    while True:
        value, chan = yield alts([c, timeout(0.8)])
        if chan is c:
            print value
        else:
            print "You're too slow."
            yield stop()



# -*- coding: utf-8 -*-
from setuptools import setup
import os

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name="twisted-csp",
      version="0.1.5",
      description="Go-style channels for Twisted",
      author="Nguyễn Tuấn Anh",
      author_email="ubolonton@gmail.com",
      url="https://github.com/ubolonton/twisted-csp",
      long_description=open('README.md').read(),
      license="EPL",

      packages=["csp", "csp.impl"],

      install_requires=[
          "twisted"
      ]
)



from csp import Channel, put, take, go, alts, wait


def produce(chan, value):
    yield wait(0.1)
    yield put(chan, value)


def main():
    chans = []
    for i in range(20):
        chan = Channel()
        go(produce(chan, i))
        chans.append(chan)

    def timeout(seconds):
        chan = Channel()
        def t():
            yield wait(seconds)
            chan.close()
        go(t())
        return chan

    chans.append(timeout(0.3))

    while True:
        value, chan = yield alts(chans)
        if value is None:
            print "time out"
            break
        else:
            print value
